import numpy as np
import sys
import os
import re
import h5py
import collections


def count(word, inf):
  pattern = r".*" + word + r".*"
  return sum(1 for line in inf if re.search(pattern, line))

def basic(line):
  line = line.split()
  if line[0] == 'GAUSSIAN':
    typ = 0
  else:
    print('Fail to recognize function type!')
  return int(typ), int(line[1]), int(line[4]), int(line[6])

def splitin3(s):
  return [i[::-1] for i in re.findall('\w{1,3}', s[::-1])[::-1]]

def center(inf, num_nuc, nline):
  pos = []
  for i in range(num_nuc):
    line = inf[nline+i].replace('CENTRE', ' ').split()
    #print(line[4:7])
    pos.append(line[4:7])
  return np.array(pos).astype(np.float)

def flatten(l):
  for el in l:
    if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
      yield from flatten(el)
    else:
      yield el

def assign(inf, item, num_prt):
  num_item = count(item, inf)
  arr = []
  for i in range(num_item):
    line = inf[nline+i].split()
    arr.extend(line[len(item.split()):])

  # to avoid no space between two number if more than 99 atoms
  if item == 'CENTRE ASSIGNMENTS':
    arr = [splitin3(s) if len(s)>3 else s for s in arr]
    #print(str(arr).count(",")+1)
    if any(isinstance(i, list) for i in arr): # check if arr is nested
      arr = list(flatten(arr))
  #print(len(arr))
  #print(num_prt)
  if len(arr) != num_prt:
    print('error: numbers in ' + item + ' not equal to number of primitives')
  return np.array(arr), num_item

def orbital(inf, nline, lines, num_orb):
  noon = []
  noe = []
  nocoef = []

  for n in range(num_orb):
    item = inf[nline].split()
    noon.append(item[7])
    noe.append(item[-1])
    nline += 1

    coefl = []
    for i in range(lines):
      coef = inf[nline+i].split()
      coefl.extend(coef)
    nocoef.append(coefl)
    nline += lines

  return np.array(noon).astype(np.float), np.array(noe).astype(np.float), np.array(nocoef).astype(np.float)


wfn = sys.argv[1]

fwfn = open(wfn, 'r')

# load the whole file
inf = fwfn.readlines()

# modify the float numbers
pattern = r'([0-9]+)D([\-|+][0-9])'
inf = [re.sub(pattern, r"\1E\2", i) for i in inf]

nline = 0

# read the title
title = inf[nline].strip()[0]
name = os.path.basename(wfn).rsplit('.',1)[0] #+ '_' + title
nline += 1

# read type of orb., num of mol orb., num of prts, num of nuc.
typ_orb, num_orb, num_prt, num_nuc = basic(inf[nline])
nline += 1

# read atom centers
pos = center(inf, num_nuc, nline)
nline += num_nuc

# read center assignments and type assignments
cassign, lines = assign(inf, 'CENTRE ASSIGNMENTS', num_prt)
cassign = cassign.astype(np.int)
nline += lines
tassign, lines = assign(inf, 'TYPE ASSIGNMENTS', num_prt)
tassign = tassign.astype(np.int)
nline += lines
exp, lines = assign(inf, 'EXPONENTS', num_prt)
exp = exp.astype(np.float)
nline += lines

# read MOs (we still use the var lines)
noon, noe, nocoef = orbital(inf, nline, lines, num_orb)

fwfn.close()


# sace to hdf5 file
with h5py.File(name + '.h5', 'w') as hf:
  hf.create_dataset("BASIC", data=np.array([typ_orb, num_orb, num_prt, num_nuc]))
  hf.create_dataset("POS",  data=pos)
  hf.create_dataset("CASSIGN",  data=cassign)
  hf.create_dataset("TASSIGN",  data=tassign)
  hf.create_dataset("EXP",  data=exp)
  hf.create_dataset("NOON",  data=noon)
  hf.create_dataset("NOE",  data=noe)
  hf.create_dataset("NOCOEF",  data=nocoef)

 
